package ca;

import glmodel.GL_Vector;

import java.util.ArrayList;

import mechanics.CellGrid;
import mechanics.Game;


public class Cell {

	private int posX;
	private int posY;
	private int posZ;
	
	private int currentState;
	private int nextState;
	private RuleBook rulebook;
	private CellGrid cellgrid;
	private boolean hasChangedState = true;
	
	private int currentGeneration = 0;
	
	private ArrayList<Integer> stateHistory;
	private String neighbourhoodHash = "";
	
	public Cell(CellGrid cellgrid, RuleBook rulebook, int posX, int posY, int posZ, int state){
		this.currentState = state;
		setX(posX);
		setY(posY);
		setZ(posZ);
		
		this.rulebook = rulebook;
		this.nextState = state;
		this.cellgrid = cellgrid;
		
		stateHistory = new ArrayList<Integer>();
	}
	
	public Cell(CellGrid cellgrid, RuleBook rulebook, int posX, int posY, int posZ){
		this( cellgrid, rulebook, posX, posY, posZ, 0);
	}

	
	/**
	 * Sets the next state as current state
	 */
	public synchronized void updateState(){
		stateHistory.set(currentGeneration, currentState);

		currentState = nextState;
		
		neighbourhoodHash  = "";
		
		currentGeneration++;
	}
	
	public int getX(){
		return posX;
	}

	public int getY(){
		return posY;
	}
	
	public boolean hasChangedState(){
		return hasChangedState;
	}
	
	public int getZ(){
		return posZ;
	}

	private void setX(int posX) {
		this.posX = posX;
	}
	private void setY(int posY) {
		this.posY = posY;
	}

	private void setZ(int posZ) {
		this.posZ = posZ;
	}
	
	@Override
	public String toString() {
		return "Cell ("+getX()+","+getY()+","+getZ()+") State: "+getState();
	}

	public int getState() {
		return currentState;
	}

	public void updateNextState() {
		// If state in history, use it
		if(currentGeneration < stateHistory.size()){
			System.out.println("Gets history from gen "+currentGeneration );
			nextState = stateHistory.get(currentGeneration);
		} else {
			// else calculate new
			nextState = rulebook.getNextState(cellgrid, posX, posY, posZ);
			stateHistory.add(nextState);
		}
		
		
		hasChangedState = (currentState != nextState);
		
		if(Game.DEBUG) System.out.println("Next state "+nextState + " x:"+posX + " y:"+posY + " z:" +posZ);
	}
	
	public void setState(int state){
		currentState = state;
	}
	
	public Integer revertStateToGeneration(int generation){
		if(generation >= currentGeneration) return currentState;
		
		currentGeneration = generation;
		/*
		// Reverting more than possible, revert max
		if( currentGeneration < generations ){
			generations = currentGeneration;
		}
		
		int stateToRevertTo = currentGeneration-generations;
		*/
		nextState = stateHistory.get(generation);
		
		return nextState;
	}

	public String getHash() {
		// If not changed, return
		if(!neighbourhoodHash.isEmpty()){
			return neighbourhoodHash;
		}
		
		// Else calculate hash
		String ret = "";
		
		for(int tempY = posY-1; tempY<=posY+1; tempY++){
			for(int tempZ = posZ-1; tempZ<=posZ+1; tempZ++){
				for(int tempX = posX-1; tempX<=posX+1; tempX++){
					if(tempX >= 0 && tempY >= 0 && tempZ >= 0 && tempX < cellgrid.getGridWidth() && tempY < cellgrid.getGridHeight() && tempZ < cellgrid.getGridDepth() ){
						Integer c = cellgrid.getCellState(tempX, tempY, tempZ);
						if(c != null){
							ret += c+"-";
						} else {
							ret += "0-";
						}
					} else {
						ret += "0-";
					}
				}
			}
		}
		
		neighbourhoodHash = ret;
		return ret;
	}

	public GL_Vector getPosition() {
		return new GL_Vector(posX, posY, posZ);
	}
	
	
	

}
