package ca;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;

import mechanics.CellGrid;
import mechanics.Game;

import ca.gui.Visualizer;
import ca.rules.HashRule;

public class RuleBook {
	
	
	private HashMap<String, HashRule> hashRules;
	private String ruleFile = "rules.txt";
	private ArrayList<Integer> possibleStates;
	private Visualizer visualizer;
	
	public RuleBook(Visualizer visualizer){
		this(visualizer, "rules.txt");
	}
	
	public RuleBook(Visualizer visualizer, String ruleFile){
		this.ruleFile = ruleFile;
		
		hashRules = new HashMap<String, HashRule>();
		possibleStates = new ArrayList<Integer>();
		this.visualizer = visualizer;
	}
	
	public RuleBook() {
		this(null);
	}
	
	/**
	 * Add a hash rule to rule book
	 * @param neighbourhood
	 * @param rule
	 */
	public void addHashRule(HashRule rule, boolean saveRulesToFile){
		//System.out.println("Add hashrule: "+neighbourhood);
		int newState = (int)rule.getNextState();
		// Set with all available states
		if(possibleStates.contains(newState) == false){
			possibleStates.add(newState);
		}
		
		// Add rule to hashmap
		hashRules.put(rule.getNeighbourhood(), rule);
		
		if(saveRulesToFile) saveRules();
	}
	
	/**
	 * Updates a rule and saves all rules
	 * @param rule
	 */
	public void updateHashRule(HashRule rule){
		//System.out.println("Update hash: "+neighbourhood);
		addHashRule(rule, true);
	}
	
	public Integer getNextState(CellGrid cellgrid, int x, int y, int z) {
		// Check if one of the hashes matches
		String hashFromCell = cellgrid.getCellHash(x, y, z);
		
		HashRule found = hashRules.get(hashFromCell);
		
		if(found != null){
			if(Game.DEBUG) System.out.println("Found: "+found);
			
			return found.getNextState();
		} else {
			// No other rules matches, check hashRules for don't-care states (x / X)
			// Should only check don't care rules, instead of everyone
			for(HashRule r : hashRules.values()){
				if(r.checkCell(hashFromCell)){
					return r.getNextState();
				}
			}
			return cellgrid.getCellState(x,y,z);
		}
	}
	
	
	public int getNumberOfNeightbourCells(CellGrid cellgrid, Cell cell) {
		int numberofNeighbourCells = 0;
		
		for(int tempX = cell.getX()-1; tempX<=cell.getX()+1; tempX++){
			for(int tempY = cell.getY()-1; tempY<=cell.getY()+1; tempY++){
				for(int tempZ = cell.getZ()-1; tempZ<=cell.getZ()+1; tempZ++){
					if(tempX >= 0 && tempY >= 0 && tempZ >= 0 && tempX < cellgrid.getGridWidth() && tempY < cellgrid.getGridHeight() && tempZ < cellgrid.getGridDepth() ){
						Integer c = cellgrid.getCellState(tempX, tempY, tempZ);
						if(c != null && c > 0 && !(tempX == cell.getX() && tempY == cell.getY() && tempZ == cell.getZ())){
							numberofNeighbourCells++;
						}
					}
				}
			}
		}
		return numberofNeighbourCells;
	}
	public void loadRules(String ruleFile){
		loadRules(ruleFile, true);
	}
	
	public void loadAdditionalRules(String filename){
		loadRules(filename, false);
	}
	
	public void loadRules(String ruleFile, boolean clearRules){
		if(clearRules) hashRules.clear();
		
		FileReader input;
		try {
			input = new FileReader(ruleFile);
			BufferedReader bufRead = new BufferedReader(input);
			
			String line;
			line = bufRead.readLine();
			
			while (line != null){
				
				String[] rule = line.split("@");
				
				if(rule.length == 3){
					addHashRule(new HashRule(rule[0], Integer.parseInt(rule[1]), rule[2]), false);
				}
				
				line = bufRead.readLine();
			}
			
			//System.out.println("Loaded "+hashRules.size()+" rules from " + ruleFile);
		} catch (FileNotFoundException e) {
			System.err.println("Could not find "+ruleFile);
		} catch (IOException e) {
			System.err.println("Could not read "+ruleFile);
		}
	}
	
	public void saveRules(String ruleFile){
		PrintWriter out = null;
		try {
			out = new PrintWriter(new FileWriter(ruleFile));
			
			for(HashRule rule : hashRules.values()){
				out.println(rule.getNeighbourhood()+"@"+rule.getNextState()+"@"+rule.getDescription());
			}
		} catch (IOException e) {
			System.err.println("Could not save "+ruleFile);
		} finally {
	      if (out != null) {
	    	  if(Game.DEBUG) System.out.println("Save rules to file "+ruleFile);
	          out.close();
	        } else {
	          System.out.println("PrintWriter not open");
	        }
		}
		
		if(visualizer != null){
			visualizer.setText("Reloading rules");
		}
	}
	
	public void saveRules(){
		saveRules(ruleFile);
	}

	public String getRuleFile(){
		return ruleFile;
	}
	
	public void setRuleFile(String ruleFile){
		this.ruleFile = ruleFile;
	}

	public HashMap<String, HashRule> getHashRules() {
		return hashRules;
	}

	public ArrayList<Integer> getPossibleStates(){
		return possibleStates;
	}


	public void removeHashRule(HashRule hashRule, boolean saveRules) {
		hashRules.remove(hashRule.getNeighbourhood());
		if(saveRules) saveRules();
	}
	
	public void removeHashRule(HashRule hashRule) {
		removeHashRule(hashRule, true);
	}
	
	@Override
	public String toString() {
		return "RuleBook "+getRuleFile();
	}

	public void loadRules() {
		loadRules(ruleFile);
	}
}
