package genetic;

import mechanics.Game;

import org.lwjgl.util.vector.Vector3f;

import ca.Cell;
import ca.CellStructures;
import ca.RuleBook;
import ca.rules.HashRule;

public class GAInstance {
	
	public CellOrder lastDirection;
	public int id;
	private Game game;

	public StateInformation state;
	
	private final int[] legalNeighborStates = { 3, 4, 8, 9, 10, 11 , 12};
	private final int[] legalTakeoverStates = { 3, 4 };
	private HashRule newestRule;
	private boolean isFinished;
	private int cellTypeToFind;
	
	public GAInstance(int id, int generation, int cellTypeToFind, RuleBook rulebook, int gridWidth, int gridHeight, int gridDepth){
		this.id = id;
		
		state = new StateInformation();
		
		state.generation = generation;
		this.cellTypeToFind = cellTypeToFind;
		
		game = new Game(null, CellStructures.BUILDNING_START, gridWidth, gridHeight, gridDepth);
		game.goForwardToGeneration(generation);
		state.position = findStartCell(game, 11);
		
	}
	
	
	public static Vector3f findStartCell(Game game, int startCellState) {
		
		Vector3f start = new Vector3f();
		
		for(Cell c : game.getCellsInSubGrid()){
			if(c.getState() == startCellState){
				start.x = c.getX();
				start.y = c.getY();
				start.z = c.getZ();
				
				System.out.println("Start: ("+start.x+","+start.y+","+start.z+")");
				return start;
			}
		}
		return start;
	}
	
	private CellOrder getDirection() {
		return CellOrder.values()[id];
	}
	

	private void moveDirection(CellOrder newDirection){
		switch(newDirection){
			case UP:
				state.position.y++;
				break;
			case DOWN:
				state.position.y--;
				break;
			case NORTH:
				state.position.z--;
				break;
			case SOUTH:
				state.position.z++;
				break;
			case EAST:
				state.position.x++;
				break;
			case WEST:
				state.position.x--;
				break;
		}
	}
	
	// Checks if current state.position is legal
	private boolean legalPosition(int x, int y, int z) {
		if( x < 0 || x >= game.getCellGrid().getSubGrid().getGridWidth()  || 
			y < 0 || y >= game.getCellGrid().getSubGrid().getGridHeight() ||
			z < 0 || z >= game.getCellGrid().getSubGrid().getGridDepth()
		   ){
			return false;
			}
		
		
		String hash = game.getCellGrid().getSubGrid().getCellHash(x, y, z);
		String[] hashArray = hash.split("-");
		
		// Check if cell in state.positions is in a state legal to take over
		boolean foundState = false;
		for(int state : legalTakeoverStates){
			if(Integer.parseInt(hashArray[hashArray.length/2]) == state){
				foundState = true;
			}
		}
		
		if(!foundState){
			return false;
		}
		
		String replacedHash = hash;
		for(int state : legalNeighborStates){
			replacedHash = replacedHash.replace(state+"-", "");
		}
		if(replacedHash.length() > 0){
			return false;
		}
		
		if(hash.contains("10-") || hash.contains("11-")){
			return true;
		}
		
		return false;
	}
	
	private boolean legalPosition() {
		return legalPosition((int)state.position.x, (int)state.position.y, (int)state.position.z);
	}

	public int getPoints(int failedPoints){
		int positionPoints = 0;
		int inactiveWireLength = game.getStateCount(10);
		int activeWireLength = game.getStateCount(11);
		
		int extraPoints = 0;
		double x=0, y=0, z=0;
		Cell lowestCell = null;
		
		// distance to all inactive lights
		for(Cell c : game.getCellsInSubGrid()){
			if(c.getState() == cellTypeToFind){
				if(lowestCell == null){
					lowestCell = c;
				} else {
					if(lowestCell.getY() > c.getY()){
						lowestCell = c;
					}
				}
			}
		}
		
		// Implement A* search here
		
		
		x = Math.abs(lowestCell.getX()-state.position.x);
		y = Math.abs(lowestCell.getY()-state.position.y);
		z = Math.abs(lowestCell.getZ()-state.position.z);
		System.out.println("("+x+","+y+","+z+")");
		
		extraPoints += x*x;
		extraPoints += y*y*y;
		extraPoints += z+z;
		
		positionPoints += extraPoints;
		extraPoints = 0;
		
		return -inactiveWireLength - activeWireLength - positionPoints - failedPoints;
	}


	public boolean tryDirection() {
		state.generation++;
		CellOrder direction = getDirection();
		
		float originalX = state.position.x;
		float originalY = state.position.y;
		float originalZ = state.position.z;
		
		moveDirection(direction);
		
		boolean wasLegal = false;
		int failedPoints = 0;
		
		if(legalPosition()){
		
			String hash = game.getCellGrid().getSubGrid().getCellHash((int)state.position.x, (int)state.position.y, (int)state.position.z);
			
			HashRule newRule = new HashRule(hash, 10, "Generation: "+state.generation+" Direction: "+direction);
			
			newestRule = newRule;
			
			wasLegal = true;
		} else {
			// reset, not valid direction
			state.position.x = originalX;
			state.position.y = originalY;
			state.position.z = originalZ;
			
			System.out.println("Not legal "+direction);
			newestRule = null;
			failedPoints = 10000;
		}
		
		
		state.points = getPoints(failedPoints);
		System.out.println("Points: "+state.points);
		//state.points = getPoints(wasLegal, direction, failedPoints);
		
		return wasLegal;
	}


	public void step() {
		game.step();
	}


	public void setState(StateInformation state) {
		this.state = state;
	}


	public HashRule getRule() {
		return newestRule;
	}


	public Game getGame() {
		return game;
	}


	public boolean isFinished() {
		return isFinished;
	}
}
