package mechanics;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import ca.Cell;
import ca.RuleBook;
import ca.gui.Visualizer;

public class CellGrid {

	private CellGrid subGrid;
	private Integer[][][] cellGrid;
	private Map<String, Cell> activeCells;
	private RuleBook rulebook;
	private int gridWidth = 0;
	private int gridHeight = 0;
	private int gridDepth = 0;
	private boolean isSubgrid = false;
	private int subCellsPerCell = 1;
	
	public CellGrid(RuleBook rulebook, int gridWidth, int gridHeight, int gridDepth, int cellsPerCell, boolean isSubgrid) {
		cellGrid = new Integer[gridWidth][gridHeight][gridDepth];
		activeCells = new HashMap<String, Cell>();
		this.rulebook = rulebook;
		
		setGridWidth(gridWidth);
		setGridHeight(gridHeight);
		setGridDepth(gridDepth);
		this.isSubgrid = isSubgrid;
		
		this.subCellsPerCell = cellsPerCell;
		
		if(isSubgrid){
			// Place out start block for the active electrical system
			//setCell(gridWidth-1, 1, gridDepth/2, 11);
		}

	}
	
	public CellGrid(RuleBook rulebook, int gridWidth, int gridHeight, int gridDepth){
		this(rulebook, gridWidth, gridHeight, gridDepth, 1, false);
	}

	
	public void createSubGrid(Visualizer visualizer, int subCellsPerCell, int gridWidth, int gridHeight, int gridDepth){
		RuleBook subRules = new RuleBook(visualizer, "subrules.txt");
		subRules.loadRules();
		
		// Extra width to place electric systems
		subGrid = new CellGrid(subRules, gridWidth*subCellsPerCell+1, gridHeight*subCellsPerCell, gridDepth*subCellsPerCell, subCellsPerCell, true);
	}
	
	public void populateNeighbourCells(int x, int y, int z) {
		Integer c = getCellState(x,y,z);
		if(c == null || c == 0) return;
		
		for(int localX=x-1; localX<=x+1; localX++){
			for(int localY=y-1; localY<=y+1; localY++){
				for(int localZ=z-1; localZ<=z+1; localZ++){
					if(localX >= 0 && localY >= 0 && localZ >= 0 && localX < getGridWidth() && localY < getGridHeight() && localZ < getGridDepth() ){
						Integer cell = getCellState(localX, localY, localZ);
						if(cell == null){
							setCell(localX, localY, localZ, 0, true);
						}
					}
				}
			}
		}
	}

	public Integer getCellState(int x, int y, int z){
		if(Game.DEBUG) System.out.println("getCell("+x+", "+y+", "+z+")");
		if(x >= 0 && y >= 0 && z >= 0 && x < getGridWidth() && y < getGridHeight() && z < getGridDepth() ){
			return cellGrid[x][y][z];
		}
		return null;
	}
	
	public void step(){
		
		if(Game.DEBUG) System.out.println("Number of cells: "+activeCells.size());
		
		String[] keys = new String[activeCells.size()];
		activeCells.keySet().toArray(keys);
		for(int index = 0; index<keys.length; index++){
			Cell c = activeCells.get(keys[index]);
			populateNeighbourCells(c.getX(), c.getY(), c.getZ());
		}

		for(Cell c : activeCells.values()){
			// Update to next state
			c.updateNextState();
		}
		
		// All cells have calculated next state, now update them!
		for(Cell c : activeCells.values()){
			// Update to next state
			c.updateState();
			cellGrid[c.getX()][c.getY()][c.getZ()] = c.getState();
			
			if(subGrid != null && isSubgrid == false && c.hasChangedState()){
				subGrid.setCellCluster(c.getX(), c.getY(), c.getZ(), c.getState());
			}
		}
		
		// Last, update subgrid
		if(subGrid != null){
			subGrid.step();
		}
	}
	
	// Updates a cluster of cells ( change of state in grid updates coresponding sub-cells)
	private void setCellCluster(int x, int y, int z, int state) {
		if(isSubgrid == false) return;
		x = x * subCellsPerCell;
		y = y * subCellsPerCell;
		z = z * subCellsPerCell;
		
		if(Game.DEBUG) System.out.println("Setting "+subCellsPerCell+"subcells for x:"+x+" y:"+y+" z:"+z);
		
		if(x >= 0 && y >= 0 && z >= 0 && x < getGridWidth() && y < getGridHeight() && z < getGridDepth() ){
			for(int localX=x; localX<x+subCellsPerCell; localX++){
				for(int localY=y; localY<y+subCellsPerCell; localY++){
					for(int localZ=z; localZ<z+subCellsPerCell; localZ++){
						if(localX >= 0 && localY >= 0 && localZ >= 0 && localX < getGridWidth() && localY < getGridHeight() && localZ < getGridDepth() ){
							setCell(localX, localY, localZ, state);
						}
					}
				}
			}
		}
	}
	
	private String planeAsString(int yValue){
		String plane = "";
		for(int tempZ = 0; tempZ<getGridDepth(); tempZ++){
			for(int tempX = 0; tempX<getGridWidth(); tempX++){
				Integer c = getCellState(tempX, yValue, tempZ);
				if(c != null){
					plane += c+" ";
				} else {
					plane += "X ";
				}
			}
			plane += "\n";
		}
		return plane;
	}

	public void printOutPlane(int yValue) {
		System.out.println(planeAsString(yValue));
	}

	public Collection<Cell> getActiveCells() {
		return activeCells.values();
	}


	public CellGrid getSubGrid() {
		return subGrid;
	}

	public boolean setCell(int x, int y, int z, Integer state, boolean updateSubgrid) {
		//System.out.println("Set cell in "+(isSubgrid?"subgrid":"grid")+" x:+"+x+" y:"+y+" z:"+z+" gridWidth:"+gridWidth);
		if(x >= 0 && y >= 0 && z >= 0 && x < getGridWidth() && y < getGridHeight() && z < getGridDepth() ){
			cellGrid[x][y][z] = state;
			if(state != null) activeCells.put(x+","+y+","+z, new Cell(this, rulebook, x, y, z, state));
			if(updateSubgrid && getSubGrid() != null) getSubGrid().setCellCluster(x, y, z, state);
			return true;
		}
		return false;
	}
	
	public boolean setCell(int x, int y, int z, Integer state) {
		return setCell(x, y, z, state, false);
	}


	public void clearCells() {
		for(Cell c : getActiveCells()){
			setCell(c.getX(), c.getY(), c.getZ(), null, false);
		}
		
		getActiveCells().clear();
		
		if(subGrid != null){
			subGrid.clearCells();
		}
	}


	public RuleBook getRuleBook() {
		return rulebook;
	}


	public boolean isSubgrid() {
		return isSubgrid;
	}

	public void setGridWidth(int gridWidth) {
		this.gridWidth = gridWidth;
	}

	public int getGridWidth() {
		return gridWidth;
	}

	public void setGridHeight(int gridHeight) {
		this.gridHeight = gridHeight;
	}

	public int getGridHeight() {
		return gridHeight;
	}

	public void setGridDepth(int gridDepth) {
		this.gridDepth = gridDepth;
	}

	public int getGridDepth() {
		return gridDepth;
	}

	public String getCellHash(int x, int y, int z) {
		return activeCells.get(x+","+y+","+z).getHash();
	}

	public void revert(int stopGeneraiton) {
		
		for(Cell c : getActiveCells()){
			setCell(c.getX(), c.getY(), c.getZ(), c.revertStateToGeneration(stopGeneraiton));
			c.updateState();

			if(subGrid != null && isSubgrid == false && c.hasChangedState()){
				subGrid.setCellCluster(c.getX(), c.getY(), c.getZ(), c.getState());
			}
		}
		
		if(!isSubgrid()) subGrid.revert(stopGeneraiton);
	}

	
}
