package mechanics;

import java.util.ArrayList;
import java.util.Collection;

import ca.CallBack;
import ca.Cell;
import ca.CellStructures;
import ca.RuleBook;
import ca.Util;
import ca.gui.Visualizer;

public class Game {
	
	private RuleBook rulebook;
	public static int gridWidth = 6;
	public static int gridHeight = 13;
	public static int gridDepth = 6;
	public static int subCellsPerCell = 3;
	public static boolean DEBUG = false;
	private ArrayList<CallBack> callbacks;
	private int generation = 0;
	private int startState = -1;
	private int generationLock = 0;
	private CellGrid cellGrid;
	private long startTime;
	private boolean structureIsHidden = false;
	
	public Game(Visualizer visualizer, int startState, RuleBook rulebook, int gridWidth, int gridHeight, int gridDepth){
		this.rulebook = rulebook;
		this.rulebook.loadRules();

		callbacks = new ArrayList<CallBack>();
		cellGrid = new CellGrid(rulebook, gridWidth, gridHeight, gridDepth);
		
		// if want subgrid:
		cellGrid.createSubGrid(visualizer, subCellsPerCell, gridWidth, gridHeight, gridDepth);
		
		String generationInt = Util.readStringFromFile("generationLock.txt");
    	if( !generationInt.isEmpty() ){
    		generationLock = Integer.parseInt(generationInt);
    	}
    	
    	if(startState >= 0){
	    	this.startState = startState;
	    	resetGeneration();
    	}
	}
	
	public Game(Visualizer visualizer, int startState, RuleBook rulebook){
		this(visualizer, startState, rulebook, gridWidth, gridHeight, gridDepth);
	}
	
	public Game(Visualizer visualizer){
		this(visualizer, -1);
	}
	
	public Game(Visualizer visualizer, int startState){
		this(visualizer, startState, new RuleBook(visualizer, "rules.txt"));
	}
	
	public Game(Visualizer visualizer, int startState, int gridWidth, int gridHeight, int gridDepth){
		this(visualizer, startState, new RuleBook(visualizer, "rules.txt"), gridWidth, gridHeight, gridDepth);
	}
	
	
	public void goForwardToGeneration(int stopGeneration){
		for(int i = generation; i<stopGeneration; i++){
			step();
		}
		generation = stopGeneration;
	}
	
	public boolean revertToGeneration(int stopGeneraiton){
		if(stopGeneraiton >= generation) return false;
		
		//System.out.println("Reverting to "+stopGeneraiton);
		//cellGrid.revert(stopGeneraiton);
		
		if(stopGeneraiton == 0) resetGeneration();
		
		generation = stopGeneraiton-1;
		
		step();
		
		for(CallBack cb : callbacks){
    		cb.methodToCallBack();
    	}
		
		return true;
	}
	
	/*
	public static void main(String[] args) {
		new SimpleGUI();
	}
	*/
	
	public void addCallBack(CallBack cb){
		callbacks.add(cb);
	}
	
	public void removeCallBack(CallBack cb){
		callbacks.remove(cb);
	}

	public void step(){
		startTime = System.currentTimeMillis();
		cellGrid.step();
		
		if(Game.DEBUG) System.out.println("Next run");
		
		generation++;

    	for(CallBack cb : callbacks){
    		cb.methodToCallBack();
    	}
    	
    	if(Game.DEBUG) cellGrid.printOutPlane(0);
    	
    	if(Game.DEBUG) System.out.println("Time: "+(System.currentTimeMillis()-startTime)+" ms");
	}
	
	public int getGeneration(){
		return generation;
	}

	private void loadStartState(int startState) {
		CellStructures cs = new CellStructures(cellGrid, rulebook);
		Cell[] cellStartState = new Cell[0];
		
		switch(startState){
		
			case CellStructures.BUILDNING_START:
				cellStartState = cs.getBuildningStart(this);
			break;
			
		}
		
		// Start cells
		for(Cell c : cellStartState){
			cellGrid.setCell(c.getX(), c.getY(), c.getZ(), c.getState(), true);
			cellGrid.populateNeighbourCells(c.getX(), c.getY(), c.getZ());
		}
		
	}

	public RuleBook getRuleBook(){
		return cellGrid.getRuleBook();
	}
	

	public RuleBook getSubRuleBook() {
		if(cellGrid.getSubGrid() != null){
			return cellGrid.getSubGrid().getRuleBook();
		} else {
			return null;
		}
	}

	public void clearCells() {
		cellGrid.clearCells();
		
		for(CallBack cb : callbacks){
    		cb.methodToCallBack();
    	}
	}

	public void resetGeneration() {
		generation = 0;
		
		if(startState >= 0){
			loadStartState(startState);
		}
		
		for(int i = 0; i < generationLock; i++){
			step();
		}
		
		if(cellGrid.getSubGrid() != null){
			cellGrid.getSubGrid().setCell(cellGrid.getSubGrid().getGridWidth(), 1, cellGrid.getSubGrid().getGridDepth()/2, 11);
		}
	}

	public void setGenerationLock(int generation) {
		generationLock = generation;
		Util.writeToFile("generationLock.txt", ""+generationLock);
	}



	public Collection<Cell> getCellsInGrid() {
		return cellGrid.getActiveCells();
	}
	
	public Collection<Cell> getCellsInSubGrid() {
		if(cellGrid.getSubGrid() != null){
			return cellGrid.getSubGrid().getActiveCells();
		} else {
			return new ArrayList<Cell>();
		}
	}



	public CellGrid getCellGrid() {
		return cellGrid;
	}

	public int getStateCount(int state) {
		int cellsWithState = 0;
		for(Cell c : getCellsInSubGrid()){
			if(c.getState() == state){
				cellsWithState++;
			}
		}
		return cellsWithState;
	}

	public boolean hasSubgrid() {
		return (cellGrid.getSubGrid()!=null);
	}

	public void toggleStructureIsHidden() {
		structureIsHidden = !structureIsHidden;
	}
	
	public boolean structureIsHidden(){
		return structureIsHidden;
	}

}
